/*-
 * Copyright (c) 2003
 *		Stepachev Andrey <aka50@mail.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/* 
 * cmd.c -- commands 
 *
 * $Id: cmd.c,v 1.10 2003/08/31 08:15:30 aka50 Exp $
 */

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "dmalloc.h"
#include "membuf.h"
#include "queue.h"

#include "ichat.h"
#include "cmd.h"
#include "gui.h"
#include "proto.h"
#include "user.h"
#include "conv_rus.h"

#define CMD_MAX_CMDARGS 5

struct spword {
	int beg;
	int end;
};

struct cmdcb{
	char *cmdname;
	int   cmdargs;
	int   argscmp[CMD_MAX_CMDARGS];
	int   (*cmdfunc)();
	char *cmdusg;
	char *cmddescr;
};

static int cmd_do_msg(const char *buf, struct spword *words, int wc);
static int cmd_do_mch(const char *buf, struct spword *words, int wc);

typedef int (*cmdfunc_t)(const char *buf, struct spword *words, int wc);


static void recalc(struct cmd_line *cmd);
static int getpos(struct cmd_line *cmd, int dir, int part);
static int splittowords(char *str, struct spword *words, int maxwrds);
static int cmd_addhist(struct cmd_line *cmd, const char *histval);

struct cmdcb cmds[];

TAILQ_HEAD(hstnd_list, hist_node);


/* Commands supported by ichat client */

static int 
cmd_do_msg(const char *buf, struct spword *words, int wc)
{
	struct membuf pkt, msg;

	mb_init(&pkt);
	mb_init(&msg);

	ich_cmd_text(&msg, ICHAT_COMMON_LINE, ICHAT_TOALL, &buf[words[0].beg]);
	ich_pkt_make(&pkt, ICHAT_TOALL, &msg);
	ich_output(&pkt);

	mb_free(&pkt);
	mb_free(&msg);

	return 0;
}

static int 
cmd_do_mch(const char *buf, struct spword *words, int wc)
{
	struct membuf pkt, msg;
	char chnl[CMD_MAX_CHNLNM + 1];
	struct user *n;
	
	int rc;

	rc = -1;

	mb_init(&pkt);
	mb_init(&msg);

	if (words[0].end - words[0].beg > CMD_MAX_CHNLNM){
		gui_put_log("MCH: Line name is to big: max = %d", CMD_MAX_CHNLNM);
		return -1;
	}

	memcpy(chnl, &buf[words[0].beg], words[0].end - words[0].beg + 1);
	chnl[words[0].end - words[0].beg + 1] = '\0';

	if (chnl[0] == '#'){
		n = find_userbynick(chnl+1);
		if (n == NULL){
			gui_put_log("MCH: Unknown nick: %s", chnl);
			rc = -1;
			goto quit;
		}
		ich_cmd_text(&msg, ICHAT_PRIVATE_LINE, n->sig, &buf[words[1].beg]);
		ich_pkt_make(&pkt, n->sig, &msg);
		ich_cmd_text(&msg, ICHAT_PRIVATE_LINE, n->sig, &buf[words[1].beg]);
		ich_rpl_parse(&msg);
	} else {
		ich_cmd_text(&msg, chnl, ICHAT_TOALL, &buf[words[1].beg]);
		ich_pkt_make(&pkt, ICHAT_TOALL, &msg);
	}


	rc = ich_output(&pkt);

 quit:
	mb_free(&pkt);
	mb_free(&msg);

	return rc;
}

static int 
cmd_do_help(const char *buf, struct spword *words, int wc)
{
	int i;

	i = 0;
	while(cmds[i].cmdname){
		gui_put_log("Usage: %s", cmds[i].cmdusg);
		gui_put_log("About: %s", cmds[i].cmddescr);
		i++;
	}

	return 0;
}

static int 
cmd_do_ulst(const char *buf, struct spword *words, int wc)
{
	return user_sys_list();
}

static int 
cmd_do_quit(const char *buf, struct spword *words, int wc)
{
	f_quit = 1;

	return 0;
}


struct cmdcb cmds[] = {
	{
		"msg", 1, {0,0,0,0,0}, (cmdfunc_t)&cmd_do_msg,
		"/MSG <text>",
		"Sends text to common channel"
	},
	{
		"mch", 2, {0,0,0,0,0}, (cmdfunc_t)&cmd_do_mch,
		"/MCH <channel> <text>",
		"Sends text to specified channel. Channel can be channel or user name prefixed by #"
	},
	{
		"help", 0, {0,0,0,0,0}, (cmdfunc_t)&cmd_do_help,
		"/HELP",
		"Prints out all commands"
	},
	{
		"ulst", 0, {0,0,0,0,0}, (cmdfunc_t)&cmd_do_ulst,
		"/ULST",
		"Lists all users"
	},
	{
		"quit", 0, {0,0,0,0,0}, (cmdfunc_t)&cmd_do_quit,
		"/quit",
		"Bail out..."
	},
	{NULL}
};


/* Command interface implementation */

static void
hist_free(struct hist_node *hn)
{
	DFREE(hn->val);
	DFREE(hn);
	return;
}


static void
recalc(struct cmd_line *cmd)
{
	int ratio, bufvis, prmptvis;

	prmptvis = cmd->prmptsz;
	bufvis = cmd->visln - prmptvis;

	/* if prompt too long... fix it... */
	if (bufvis < 0 || bufvis >> 2 < prmptvis){
		ratio = cmd->visln >> 2;
		bufvis = ratio << 2;
		prmptvis = cmd->visln - bufvis;
	}

	cmd->prmptsz = prmptvis;

	/* if cursor is out of the right edge ... */
	if (cmd->bufpos > cmd->bufstrt + bufvis)
		cmd->bufstrt = cmd->bufpos - bufvis + 1;

	if (cmd->bufpos < cmd->bufstrt)
		cmd->bufstrt = cmd->bufpos;

	cmd->buf[cmd->bufsz] = '\0';

	return;
}

static int
getpos(struct cmd_line *cmd, int dir, int part)
{
	register int newpos;

	newpos = cmd->bufpos;

	switch(part){
	case CMD_PRT_CHAR:
		newpos += (dir == CMD_DIR_FWD)?1:-1;
		break;
	case CMD_PRT_END:
	        newpos = (dir == CMD_DIR_FWD)?cmd->bufsz: 0;
		break;
	default:
		return -1;
	}

	if (newpos < 0)
		newpos = 0;
	if (newpos > cmd->bufsz)
		newpos = cmd->bufsz;
	return newpos;
}

static int
splittowords(char *str, struct spword *words, int maxwrds)
{
	int i;
	int wc;
	int f_inwrd;

	wc = 0;
	i = 0;

	if (maxwrds < 1)
		return 0;

	/* skip all white space at the beginnig of the line */
	while(str[i] != '\0' && isspace(str[i]))
		i++;

	if (str[i] == '\0')
		return wc; /* 0 */


	f_inwrd = 1;
	words[wc].beg = i;

	while(str[i] != '\0'){
		register int f_isspace, len;

		if (str[i] == '\\' && str[i+1] != 0){
			len = strlen(str + i + 1);
			memcpy(&str[i], &str[i+1], len);
			switch(str[i]){
			case 'n':
				str[i] = '\n';
				break;
			case 'b':
				str[i] = '\b';
				break;
			case 's':
				str[i] = ' ';
				break;
			default:
				break;
			}
			f_isspace = 0;
		} else 
			f_isspace = isspace(str[i]);

		if (f_inwrd && wc == maxwrds - 1){
			words[wc++].end = strlen(str);
			f_inwrd = 0;
			break;
		}

		if (f_isspace && f_inwrd){
			words[wc++].end = i-1;
			f_inwrd = 0;
		} else if (!f_isspace && !f_inwrd){
			words[wc].beg = i;
			f_inwrd = 1;
		}
		i++;
	}

	if (f_inwrd)
		words[wc++].end = i - 1;

	return wc;
}

#if 0
static int
lookupcurwrd(struct spword *words, int wc, int pos){
	register int i;

	if (wc == 0)
		return -1;

	for (i = 0; i < wc; i++){
		if (pos < words[i].beg)
			break;
	}

	return i-1;
}
#endif 

struct cmd_line *
cmd_new(int isize, const char *prompt)
{
	struct cmd_line *cmd;
	cmd = DTMALLOC(struct cmd_line);
	if (cmd == NULL)
		return cmd;

	memset(cmd, 0, sizeof(struct cmd_line));
	strncpy(cmd->prmpt, prompt, CMD_MAX_PRMPT);
	cmd->prmptsz = strlen(prompt);
	TAILQ_INIT(&cmd->hist);
	return cmd;
}

void
cmd_free(struct cmd_line *cmd)
{
	struct hist_node *hn;

	while(!TAILQ_EMPTY(&cmd->hist)){
		hn = TAILQ_FIRST(&cmd->hist);
		TAILQ_REMOVE(&cmd->hist, hn, link);
		hist_free(hn);
	}
	if (cmd)
		DFREE(cmd);
	return;
}

int
cmd_clear(struct cmd_line *cmd, int upto)
{
	if (upto >= CMD_MAX_BUFFER)
		upto = 0;
	memset(cmd->buf + upto, 0, CMD_MAX_BUFFER - upto);
	cmd->bufpos = upto;
	cmd->bufstrt = 0;
	cmd->bufsz = upto;
	cmd->buf[cmd->bufpos] = '\0';
	recalc(cmd);
	return 0;
}

int
cmd_setvis(struct cmd_line *cmd, int size)
{
	cmd->visln = size;
	recalc(cmd);
	return 0;
}

char *
cmd_getbuf(struct cmd_line *cmd)
{
	return cmd->buf + cmd->bufstrt;
}

int
cmd_getbufsz(struct cmd_line *cmd)
{
	return cmd->bufsz - cmd->bufstrt;
}

char *
cmd_getprmpt(struct cmd_line *cmd)
{
	return cmd->prmpt;
}

int
cmd_getprmptsz(struct cmd_line *cmd)
{
	return cmd->prmptsz;
}

int
cmd_getcurpos(struct cmd_line *cmd)
{
	return cmd->bufpos - cmd->bufstrt + cmd->prmptsz;
}

int
cmd_putch(struct cmd_line *cmd, char ch)
{
	if (cmd->bufsz > CMD_MAX_BUFFER)
		return -1;

	if (cmd->bufpos < cmd->bufsz)
		memcpy(cmd->buf + cmd->bufpos + 1, cmd->buf + cmd->bufpos , cmd->bufsz - cmd->bufpos + 1); 

	/* advance position */
	cmd->bufsz++;
	cmd->buf[cmd->bufpos++] = ch;
	cmd->buf[cmd->bufsz] = '\0';
	recalc(cmd);
	return 0;
}


int
cmd_kill(struct cmd_line *cmd, int dir, int part)
{
	register int newpos, deflsz = 0;
	newpos = getpos(cmd, dir, part);
	if (newpos < 0)
		return -1;

	if (newpos > cmd->bufpos){
		deflsz = newpos - cmd->bufpos;
		memcpy(cmd->buf + cmd->bufpos, cmd->buf + newpos, cmd->bufsz - newpos);
	} else {
		deflsz = cmd->bufpos - newpos;
		memcpy(cmd->buf + newpos, cmd->buf + cmd->bufpos, cmd->bufsz - cmd->bufpos);
		cmd->bufpos = newpos;
	}
	cmd->bufsz -= deflsz;
	recalc(cmd);
	return 0;
}

int
cmd_mvcurs(struct cmd_line *cmd, int dir, int part)
{
	register int newpos;
	newpos = getpos(cmd, dir, part);
	if (newpos < 0)
		return -1;
	cmd->bufpos = newpos;
	recalc(cmd);
	return 0;
}

int
cmd_execute(struct cmd_line *cmd)
{
	struct spword cmdwrds[CMD_MAX_CMDARGS];
	char cmdname[CMD_MAX_CMD + 1];
	int wc, cc, rc, lp;
	struct membuf pkt, msg;
	
	char *wbuf;

	cc = -1;
	lp = 0;

	mb_init(&pkt);
	mb_init(&msg);

	memset(cmdwrds, 0, sizeof(cmdwrds));

	if (cmd->bufsz < 1)
		return 0;

	wbuf = strdup(cmd->buf);
	conv_koiwin(wbuf);

	wc = splittowords(wbuf, cmdwrds, CMD_MAX_CMDARGS);
	if (wc < 1)
		return 0;

	if (wbuf[cmdwrds[0].beg] != '/'){
		ich_cmd_text(&msg, ICHAT_COMMON_LINE, ICHAT_TOALL, wbuf);
		ich_pkt_make(&pkt, ICHAT_TOALL, &msg);
		ich_output(&pkt);
		goto done;
	}

	if (cmdwrds[0].end - cmdwrds[0].beg > CMD_MAX_CMD 
	    || cmdwrds[0].beg == cmdwrds[0].end)
		goto fail;

	memcpy(cmdname, &(wbuf[cmdwrds[0].beg+1]), cmdwrds[0].end - cmdwrds[0].beg);
	cmdname[cmdwrds[0].end - cmdwrds[0].beg] = '\0';

	rc = -1;
	cc = 0;
	while(cmds[cc].cmdname){
		if (!strcmp(cmdname, cmds[cc].cmdname)){
			wc--;
			if (wc < cmds[cc].cmdargs){
				gui_put_log("Insufficient args");
				break;
			}
			rc = (*cmds[cc].cmdfunc)(wbuf, &cmdwrds[1], wc);
			if (cmds[cc].cmdargs > 0)
				lp = cmdwrds[cmds[cc].cmdargs - 1].end+2;
			break;
		}
		cc++;
	}

	if (rc != 0)
		goto fail;

 done:
	cmd_clear(cmd, lp);
	cmd_addhist(cmd, cmd->buf);

	free(wbuf);
	mb_free(&pkt);
	mb_free(&msg);

	return 0;

 fail:
	if (cc < 0 || cmds[cc].cmdname == NULL)
		gui_put_log("Unknown command");
	else {
		gui_put_log("Usage: %s", cmds[cc].cmdusg);
		gui_put_log("About: %s", cmds[cc].cmddescr);
	}

	mb_free(&pkt);
	mb_free(&msg);
	
	return -1;
}


static int 
cmd_addhist(struct cmd_line *cmd, const char *histval)
{
	struct hist_node *hn, *hn2;

	if (strlen(histval) == 0)
		return 0;

	TAILQ_FOREACH(hn, &cmd->hist, link)
		if (!strcmp(histval, hn->val))
			break;
		
	if (hn != NULL)
		return 0;

	hn = DTMALLOC(struct hist_node);
	if (hn == NULL)
		return -1;
	
	hn->val = strdup(histval);

	if (cmd->histsz > CMD_MAX_HISTSZ){
		hn2= TAILQ_LAST(&cmd->hist, hstnd_list);
		TAILQ_REMOVE(&cmd->hist, hn2, link);
		cmd->histsz--;
		hist_free(hn2);
	}
	cmd->histsz++;
	cmd->curhn = hn;
	TAILQ_INSERT_HEAD(&cmd->hist, hn, link);
	return 0;
}

int
cmd_setnexthist(struct cmd_line *cmd, int dir)
{
	struct hist_node *hn;

	/* set up current node */
	if (cmd->curhn == NULL)
		return 0;

	if (dir == 1)
		hn = TAILQ_NEXT(cmd->curhn, link);
	else
		hn = TAILQ_PREV(cmd->curhn, hstnd_list,  link);

	if (hn != NULL)
		cmd->curhn = hn;

	strncpy(cmd->buf, cmd->curhn->val, CMD_MAX_BUFFER - 1);
	cmd_clear(cmd, strlen(cmd->buf));

	return 0;
}

#if 0
int
test()
{
	struct spword words[10];
	int wc, i, wp;

	gui_put_log("In buffer: \"%s\"", cmd->buf);
	wc = splittowords(cmd->buf, words, 10);
	for(i = 0; i < wc; i++)
		gui_put_log("Word %d: from %d to %d", i, words[i].beg, words[i].end);
	
	wp = lookupcurwrd(words, wc, cmd->bufpos);
	if (wp >= 0)
		gui_put_log("Current word at position %d is %d", cmd->bufpos, wp);

	return 0;
}
#endif
