/*-
 * Copyright (c) 2003
 *		Stepachev Andrey <aka50@mail.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/* 
 * main.c -- main module for iChat client
 *
 * $Id: main.c,v 1.13 2003/08/15 21:02:29 aka50 Exp $
 */

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/socket.h>

#ifdef __linux__
#include <sys/time.h>
#include <sys/select.h>
#endif

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "membuf.h"
#include "strops.h"

#include "ichat.h"
#include "gui.h"
#include "proto.h"
#include "user.h"
#include "conv_rus.h"

static char *pname = NULL;
static struct sockaddr_in srvaddr;
static struct sockaddr_in bndaddr;

char *nickname = NULL;
char *login = NULL;
char *signature = NULL;
int   status = 0;
int   wsock = -1;
int   loglevel = 0;
int   f_quit = 0;

struct membuf obuf;
struct membuf ibuf;


/* flags */
static int f_havebndaddr = 0;

static void usage(void);
static void catch_sig(int sig);
static int  fill_ip(struct sockaddr_in *ipno, char *ip);
static int  ich_connsock(struct sockaddr_in *saddr, struct sockaddr_in *daddr);
static void do_loop(void);


static void 
usage()
{
	fprintf(stderr, "IChat client\n");
	fprintf(stderr, " * Andrey Stepachev <aka50@mail.ru>\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "usage: %s [-h] -n name -s addr[:port] [-b addr] \n", pname);

	fprintf(stderr, "	-b 	address to connect from\n");
	fprintf(stderr, "	-h 	this usefull message\n");
	fprintf(stderr, "	-n 	nickname\n");	
	fprintf(stderr, "	-i 	signature\n");	
	fprintf(stderr, "	-s 	server IP address and port\n");
	fprintf(stderr, "	-w 	don't convert to koi8\n");
	fprintf(stderr, "\n");
	exit(1);
}

static void
catch_sig(int sig)
{
	f_quit = 1;
	return;
}

void 
ich_log(int llvl, char *fmt, ...)
{
	va_list ap;
	static char logbuf[ICH_LOG_BUFSIZE + 1];
	static char *dbgstr[ICHAT_LOG_MAX + 1] = {
		"crit",
		"info",
		"info",
		"dbg1",
		"dbg2",
		"dbg3"
	};

	if(loglevel <= llvl) 
		return;
  
	va_start(ap, fmt);
	vsnprintf(logbuf, ICH_LOG_BUFSIZE - 1, fmt, ap);
	va_end(ap);

	if (gui_isactive())
		gui_put_log("[%s] %s", dbgstr[llvl] ,logbuf);
	else 
		fprintf(stdout, "[%s] %s\n", dbgstr[llvl] ,logbuf);
	return;
}

static int
fill_ip(struct sockaddr_in *ipno, char *ip)
{
	struct hostent *host;
	char *p = NULL;

	ipno->sin_family = AF_INET;
	if (!strncmp(ip,"any",strlen(ip))) {
		ipno->sin_addr.s_addr = 0;
	} else {
		p = strchr(ip, ':');
		if (p) {
			*p++ = '\0';
			ipno->sin_port = htons(atoi(p));
			if (ipno->sin_port == 0)
				return EINVAL;
		}

		if (inet_aton(ip, &(ipno->sin_addr)) < 0){
			if ((host = gethostbyname(ip)) != NULL){
				memcpy(&ipno->sin_addr, host->h_addr, sizeof(struct in_addr));
			}else{
				return EINVAL;
			}
		}
	}
	return 0;
}


static int
ich_connsock(struct sockaddr_in *saddr, struct sockaddr_in *daddr)
{
	int fd=-1;
	int rc, len, mode;

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd < 0)
		return -1;

	if (saddr != NULL){
		len=sizeof(struct sockaddr_in);
		rc = bind(fd, (struct sockaddr *)saddr, len);
		if (rc < 0)
			return -1;
	}

	rc = connect(fd, (struct sockaddr *)daddr, sizeof(struct sockaddr_in));
	if (rc != 0)
		return -1;

	mode = fcntl(fd, F_GETFL, NULL);
#ifdef USE_FIONBIO_IOCTL
	len = 1;
	ioctl(fd, FIONBIO, (char *)&len);
#else
	fcntl(fd, F_SETFL, mode | O_NONBLOCK);
#endif
	return fd;
}


int
ich_output(struct membuf *mb)
{
	return mb_append_mb(&obuf, mb);
}

static void
do_loop()
{
	fd_set rfds, wfds;
	struct membuf msgbuf;
	struct timeval tv;

	time_t ltime, lrefresh = 0;

	char buf[1024];

	int rs, rc;
	int f_connected = 0;

	mb_init(&obuf);
	mb_init(&ibuf);
	mb_init(&msgbuf);

	ich_cmd_disconnect(&msgbuf, ICHAT_COMMON_LINE);
	ich_pkt_make(&obuf, ICHAT_TOALL, &msgbuf);
	ich_cmd_connect(&msgbuf, ICHAT_COMMON_LINE, ICHAT_TOALL, "Hello there!", status);
	ich_pkt_make(&obuf, ICHAT_TOALL, &msgbuf);
	ich_cmd_refresh(&msgbuf, ICHAT_COMMON_LINE, ICHAT_TOALL, "Hello there!", status);
	ich_pkt_make(&obuf, ICHAT_TOALL, &msgbuf);

	while (f_quit == 0 || f_connected == 1){
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);

		if (mbtosz(&obuf) > 0)
			FD_SET(wsock, &wfds);

		if (f_connected == 1){
			if (!f_quit)
				FD_SET(0, &rfds);
			FD_SET(wsock, &rfds);
			if (f_quit == 1){
				ich_cmd_disconnect(&msgbuf, ICHAT_COMMON_LINE);
				ich_pkt_make(&obuf, ICHAT_TOALL, &msgbuf);
			}
		}

		tv.tv_sec = 1;
		tv.tv_usec = 0;

		rs = select(FD_SETSIZE, &rfds, &wfds, NULL, &tv);

		ltime = time(NULL);
		if (f_connected && ((ltime - lrefresh) > ICHAT_RFR_TIMEO)){
			ich_cmd_refresh(&msgbuf, ICHAT_TOALL, ICHAT_TOALL, "Hello there!", 0);
			ich_pkt_make(&obuf, ICHAT_TOALL, &msgbuf);
			lrefresh = time(NULL);
			user_lst_update();
		}


		if (rs < 1)
			continue;

		/* Some data from server... */
		if (FD_ISSET(wsock, &rfds)){
			for (;;) {
				rc = recv(wsock, buf, 1024, 0);
				if (rc == 0){
					ich_log(ICHAT_LOG_CRIT, "Connection closed by server");
					goto done;
				} else if (rc < 0){
					if (errno == EAGAIN)
						break;
					ich_log(ICHAT_LOG_CRIT, "Recv error: %s", strerror(errno));
					goto done;
				}
				mb_append(&ibuf, buf, rc);
			}
			while (mbtosz(&ibuf) > 0){
				rc = ich_pkt_parse(&ibuf, &msgbuf);
				if (rc < 0)
					goto done;
				if (rc == 0)
					break;
				if (ich_rpl_parse(&msgbuf) != 0)
					goto done;
			};
		}


		if (FD_ISSET(wsock, &wfds)){
			rc = send(wsock, mbtod(&obuf, char *), mbtosz(&obuf), 0);
			if (rc > 0)
				mb_consume(&obuf, rc);
		}

		if (mbtosz(&obuf) == 0){
			if (f_connected == 0){
				ich_log(ICHAT_LOG_VERBOSE, "Connected...");
				f_connected = 1;
			} if (f_quit == 1) {
				ich_log(ICHAT_LOG_VERBOSE, "Disconnected...");
				f_connected = 0;
			}
		}

		/* User types some keys... listen him... */
		if (FD_ISSET(0, &rfds)){
			rc = gui_handle_input(&obuf);
			if (rc != 0)
				f_quit = 1;
		}
	}

done:
	if (errno != 0){
		perror("syserr");
	}

	mb_free(&obuf);
	mb_free(&ibuf);
	mb_free(&msgbuf);
	return;
}

int
main(int argc, char **argv)
{
	extern char *optarg;
  
	int c;

	memset(&srvaddr, 0, sizeof(struct sockaddr_in));
	memset(&bndaddr, 0, sizeof(struct sockaddr_in));

/*  	loglevel = ICHAT_LOG_DEBUG;  */
  	loglevel = ICHAT_LOG_DEBUG2;  
/*   	loglevel = ICHAT_LOG_MAX;   */

	signature = strdup("2.1.1.1/test");
	login = strdup("*****");

	if (strrchr(argv[0],'/') == NULL){
		pname = argv[0];
	} else {
		pname = strrchr(argv[0],'/')+1;
	}

	while ((c = getopt(argc,argv,"hwn:b:s:i:")) > 0){
		switch (c)
		{
		case 'w':
			conv_ignore = 1;
			break;
		case 'n':
			if (nickname)
				free(nickname);
			nickname = strdup(optarg);
			break;
		case 'i':
			if (signature)
				free(signature);
			signature = strdup(optarg);
			break;
		case 'b':
			if (fill_ip(&bndaddr, optarg) != 0){
				ich_log(ICHAT_LOG_CRIT, "Invalid bind IP addr: %s", optarg);
				usage();
			}
			bndaddr.sin_port = 0; /* XXX: ignore specified port... */
			f_havebndaddr = 1;
			break;
		case 's':
			if (fill_ip(&srvaddr, optarg) != 0){
				ich_log(ICHAT_LOG_CRIT, "Invalid server IP addr: %s", optarg);
				usage();
			}
			if (srvaddr.sin_port == 0)
				srvaddr.sin_port = htons(ICHAT_DFLT_PORT);
			break;
		case 'h':
		default:
			usage();
			goto fail;
		}
	}

	if (nickname == NULL || srvaddr.sin_port == 0){
		ich_log(ICHAT_LOG_CRIT, "User name and server address are required!");
		usage();
	}

	gui_init();
	user_init();

	ich_log(ICHAT_LOG_NORMAL, "IChat v1.0: %s", nickname);

	wsock = ich_connsock(f_havebndaddr?&bndaddr:NULL, &srvaddr);
	if (wsock < 0){
		ich_log(ICHAT_LOG_CRIT, "Can't connect: %s", strerror(errno));
		goto fail;
	}

	signal( SIGQUIT, catch_sig );
	signal( SIGINT, catch_sig );

	do_loop();

	close(wsock);
fail:
	ich_log(ICHAT_LOG_NORMAL, "Bye...");

	user_fini();
	gui_fini();

	if (nickname)
		free(nickname);
	if (signature)
		free(signature);
	if (login)
		free(login);
	return 0;
}
