/*
 * membuf.c -- Memory buffers operations.
 *
 * $Id: membuf.c,v 1.4 2003/07/08 05:30:12 aka50 Exp $
 */
#include <sys/types.h>

#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "membuf.h"

#ifdef __linux__
#ifndef QUAD_MAX
#ifdef LONG_LONG_MAX
#define QUAD_MAX LONG_LONG_MAX
#else
#define QUAD_MAX 0xffffffffffffffff 
#endif
#endif /* QUAD_MAX */
#endif /* __linux__ */

int membuf_balance=0;

#define MEMBUF_PAGE(pshift, size) ((((size)>>(pshift))+1)<<(pshift))
#define MEMBUF_CHECK(mb)  			\
	if ((mb)->mb_magic != MEMBUF_MAGIC) 	\
    		mb_reinit((mb),1);			

#ifndef ERRMSG
#define ERRMSG(format, args...) fprintf(stderr, format ,## args)
#endif /* ERRMSG */

static int mb_sprintf_int(struct membuf *mb, int reset, const char *fmt,  va_list ap);

int
mb_init(struct membuf *mb)
{
	return mb_reinit(mb, 1);
}

int
mb_reinit(struct membuf *mb, int def)
{
	register size_t rsrv = 1;
	register size_t pshft = MEMBUF_DEFAULT_PSHFT;

	/* save previos settings */
	if ( mb->mb_magic == MEMBUF_MAGIC){
		if (mb->mb_pdata != NULL){
			free(mb->mb_pdata);
			membuf_balance--;
		}
		if (def == 0){
			rsrv = mb->mb_rsrv;
			pshft = mb->mb_pshft;
		}
	} 

	memset(mb,0,sizeof(struct membuf));
	mb->mb_magic = MEMBUF_MAGIC;
	mb->mb_pshft = pshft;
	mb->mb_rsrv  = rsrv;
	return MEM_OK;
}

void
mb_free(struct membuf *mb)
{
	if ( mb->mb_magic == MEMBUF_MAGIC &&
	     mb->mb_pdata != NULL){
		free(mb->mb_pdata);
		membuf_balance--;
	}
	memset(mb, 0, sizeof(struct membuf));
	return;
}

int  
mb_set_pshft(struct membuf *mb, size_t shift)
{
	MEMBUF_CHECK(mb);
	mb->mb_pshft = shift;
	return 0;
}

int
mb_set_rsrv(struct membuf *mb, size_t size)
{
	MEMBUF_CHECK(mb);
	mb->mb_rsrv = size;
	return mb_resize(mb, mb->mb_size);
}

int
mb_append_rsrv(struct membuf *mb, size_t size)
{
	register int rsz;
	MEMBUF_CHECK(mb);

	rsz = (size > mb->mb_rsrv)?mb->mb_rsrv:size;
	mb->mb_size += rsz;
	return mb_resize(mb, mb->mb_size);
}

int
mb_reset(struct membuf *mb, size_t size)
{
	mb_reinit(mb, 0);
	mb->mb_psize = MEMBUF_PAGE(mb->mb_pshft, size + mb->mb_rsrv);
	mb->mb_pdata = malloc(mb->mb_psize);
	if (mb->mb_pdata == NULL){
		mb_reinit(mb, 0);
		return MEM_ERRNO;
	}
	membuf_balance++;
	mb->mb_fb=mb->mb_pdata;
	mb->mb_size=size;
	if (mb->mb_rsrv > 0)
		mb->mb_pdata[0]=0; 
	return MEM_OK;
}

int
mb_put(struct membuf *mb, const void *data, size_t size)
{
	int err;

	MEMBUF_CHECK(mb);

	if (mbtosz(mb) != 0){
		err = mb_reset(mb, 0);
		if (err != 0)
			return err;
	}

	if (size == 0)
		return MEM_OK;

	return mb_append(mb, data, size);
}

int
mb_append(struct membuf *mb, const void *data, size_t size)
{
	register int err;
	register int osz;

	MEMBUF_CHECK(mb);

	if (size == 0)
		return MEM_OK;

	osz = mb->mb_size;
	err = mb_resize(mb, size+mb->mb_size);
	if (err != 0)
		return err;

	memcpy(mb->mb_fb + osz, data, size);
	mb->mb_fb[mb->mb_size]='\0';
	return MEM_OK;
}

int
mb_consume(struct membuf *mb, size_t size)
{
	if (size == 0)
		return 0; /* nothing to do */

	MEMBUF_CHECK(mb);
	if (mb->mb_size < size){
		ERRMSG("mb_consume: can't consume %d bytes: in buffer %d bytes\n",
		       size, mb->mb_size);
		return MEM_EINVAL;
	}
	mb->mb_fb   += size;
	mb->mb_size -= size;
	return MEM_OK;
}

int
mb_consume_to_mb(struct membuf *mb_src, struct membuf *mb_dst, size_t size)
{
	if (size == 0)
		return 0; /* nothing to do */

	MEMBUF_CHECK(mb_src);
	if (mb_src->mb_size < size || mb_src->mb_size == 0){
		ERRMSG("mb_consume: can't consume %d bytes: in buffer %d bytes\n",
		       size, mb_src->mb_size);
		return MEM_EINVAL;
	}

	mb_reinit(mb_dst, 0);
	mb_append(mb_dst, mb_src->mb_fb, size);
	mb_src->mb_fb += size;
	mb_src->mb_size -= size;
	return 0;
}

int
mb_resize(struct membuf *mb, size_t size)
{
	register int   npsz, nsz, csz, odsz, err;
	register char *pd;

	MEMBUF_CHECK(mb);

	if (mb->mb_pdata == NULL || size == 0){
		/* buffer not used yet */
		err = mb_reset(mb, size);
		if (err != 0)
			return err;
		return MEM_OK;
	}

	/* 
	 * step 1: Calculate old data size, consumed
	 *	   offset, new size and new page size
	 */
	odsz = (size > mb->mb_size)?mb->mb_size:size;
	csz = mb->mb_fb - mb->mb_pdata;
	nsz  = size + mb->mb_rsrv + csz;
	npsz = MEMBUF_PAGE(mb->mb_pshft, nsz);


	/* 
	 * step 2: Check for current page size, if
	 *  	   enought, simple set new size to 
	 *         desired.
	 */
	if (npsz == mb->mb_psize){
		mb->mb_size = size;
		if(mb->mb_rsrv > 0)
			mb->mb_fb[size] = '\0';
		return MEM_OK;
	}
	

	/*
	 * step 3: Check, if possible to move data, 
	 * 	   without reallocation page buffer,
	 *         move data 
	 */
	nsz  = size + mb->mb_rsrv;
	npsz = MEMBUF_PAGE(mb->mb_pshft, nsz);

	if (npsz == mb->mb_psize){
		memcpy(mb->mb_pdata, mb->mb_fb, odsz);
		if(mb->mb_rsrv > 0)
			mb->mb_pdata[odsz]='\0';
		mb->mb_fb = mb->mb_pdata;
		mb->mb_size = size;
		return MEM_OK;
	}

	/* 
	 * step 4: All previos checks failed.We are really
	 * 	   need buffer with different page size 
	 */
	pd = malloc(npsz);
	if (pd == NULL)
		return MEM_ERRNO;
	membuf_balance++;
    
	/* copy old data */
	if (odsz > 0)
		memcpy(pd, mb->mb_fb, odsz);
	free(mb->mb_pdata);
	membuf_balance--;

	mb->mb_pdata = pd;
	mb->mb_psize = npsz;
	mb->mb_fb=mb->mb_pdata;
	mb->mb_size = size;

	/* safe nulls */
	if(mb->mb_rsrv > 0){
		mb->mb_fb[odsz] = '\0';
		mb->mb_fb[size] = '\0';
	}

	return MEM_OK;
}


int
mb_exchange_mb(struct membuf *mb1, struct membuf *mb2)
{
	struct membuf tmb;

	MEMBUF_CHECK(mb1);
	MEMBUF_CHECK(mb2);
    
	memcpy(&tmb, mb1, sizeof(struct membuf));
	memcpy(mb1, mb2, sizeof(struct membuf));
	memcpy(mb2, &tmb, sizeof(struct membuf));

	return 0;
}


/*
 * Code from FreeBSD 4.5 libc
 *
 * Macros for converting digits to letters and vice versa
 */
#define to_digit(c)     ((c) - '0')
#define is_digit(c)     ((unsigned)to_digit(c) <= 9)
#define to_char(n)      ((n) + '0')

#define FL_LONGINT	0x1
#define FL_QUADINT	0x2
#define FL_SHRTINT	0x4
#define FL_ZEROPAD	0x8
#define FL_ALTNOTA	0x10
#define FL_HAVEPREC	0x20
#define FL_RJUSTIFY	0x40
#define FL_EXACT	0x80 /* '=' flags define to exact length in prec */

/*
 * To extend shorts properly, we need both signed and unsigned
 * argument extraction methods.
 */
#define SARG()							\
        (flags&FL_LONGINT ? va_arg(ap, long) :			\
            flags&FL_SHRTINT ? (long)(short)va_arg(ap, int) :	\
            (long)va_arg(ap, int))

#define UARG()								\
        (flags&FL_LONGINT ? va_arg(ap, u_long) :			\
            flags&FL_SHRTINT ? (u_long)(u_short)va_arg(ap, int) :	\
            (u_long)va_arg(ap, u_int))


#define u_long_t unsigned long
#define u_quad_t unsigned long long

static char *xdigs_uc = "0123456789ABCDEF";
static char *xdigs_lc = "0123456789abcdef";

/*
 * Convert an unsigned long to ASCII for printf purposes, returning
 * a pointer to the first character of the string representation.
 * Hex numbers use the given digits.
 */
static char *
__ultoa(u_long val, char *endp, int base, char *xdigs)
{
        register char *cp = endp;
        register long sval;

        /*
         * Handle the three cases separately, in the hope of getting
         * better/faster code.
         */
        switch (base) {
        case 10:
                if (val < 10) { /* many numbers are 1 digit */
                        *--cp = to_char(val);
                        return (cp);
                }
                /*
                 * On many machines, unsigned arithmetic is harder than
                 * signed arithmetic, so we do at most one unsigned mod and
                 * divide; this is sufficient to reduce the range of
                 * the incoming value to where signed arithmetic works.
                 */
                if (val > LONG_MAX) {
                        *--cp = to_char(val % 10);
                        sval = val / 10;
                } else
                        sval = val;
                do {
                        *--cp = to_char(sval % 10);
                        sval /= 10;
                } while (sval != 0);
                break;

        case 16:
                do {
                        *--cp = xdigs[val & 15];
                        val >>= 4;
                } while (val);
                break;

        default:                        /* oops */
                abort();
        }
        return (cp);
}

/* Identical to __ultoa, but for quads. */
static char *
__uqtoa(u_quad_t val, char *endp, int base, char *xdigs)
{
        char *cp = endp;
        quad_t sval;

        /* quick test for small values; __ultoa is typically much faster */
        /* (perhaps instead we should run until small, then call __ultoa?) */
        if (val <= ULONG_MAX)
                return (__ultoa((u_long)val, endp, base, xdigs));
        switch (base) {
        case 10:
                if (val < 10) {
                        *--cp = to_char(val % 10);
                        return (cp);
                }
                if (val > QUAD_MAX) {
                        *--cp = to_char(val % 10);
                        sval = val / 10;
                } else
                        sval = val;
                do {
                        *--cp = to_char(sval % 10);
                        sval /= 10;
                } while (sval != 0);
                break;

        case 16:
                do {
                        *--cp = xdigs[val & 15];
                        val >>= 4;
                } while (val);
                break;

        default:
                abort();
        }
        return (cp);
}



int
mb_append_strfmt(struct membuf *mb, const char *fmt, ...)
{
	int err;
	va_list ap;

	MEMBUF_CHECK(mb);

	va_start(ap, fmt);
	err = mb_sprintf_int(mb, 0, fmt, ap);
	va_end(ap);
	return err;
}

int
mb_sprintf(struct membuf *mb, const char *fmt, ...)
{
	int err;
	va_list ap;

	MEMBUF_CHECK(mb);

	va_start(ap, fmt);
	err = mb_sprintf_int(mb, 1, fmt, ap);
	va_end(ap);
	return err;
}

int
mb_append_vstrfmt(struct membuf *mb, const char *fmt, va_list ap)
{
	MEMBUF_CHECK(mb);

	return mb_sprintf_int(mb, 0, fmt, ap);
}

int
mb_vsprintf(struct membuf *mb, const char *fmt, va_list ap)
{
	MEMBUF_CHECK(mb);
	return mb_sprintf_int(mb, 1, fmt, ap);
}


/* 
 * Light version of snprintf 
 */
static int
mb_sprintf_int(struct membuf *mb, int reset, const char *fmt,  va_list ap)
{
#define MB_MAXPRNPARAM 128
	char sprn[MB_MAXPRNPARAM+1];
	char *pfmt;
	int  err, base;

	int  flags;
	char sign;

	char *addptr;
	int   addsz;
	int   prec;

	char 	*hextable;
	u_long_t ulval;
	u_quad_t uqval;
	char     chval;

	if (reset)
		mb_resize(mb, 0);

	while (*fmt) {
		/* check for param */
		pfmt = strchr(fmt, '%');
		if (pfmt != NULL){
			mb_append(mb, fmt, pfmt-fmt);
			fmt = pfmt;
		}else{
			mb_append(mb, fmt, strlen(fmt));
			break;
		}
		/* check for ``%%'', and skip it */

		sprn[MB_MAXPRNPARAM]='\0';
		addptr = sprn + MB_MAXPRNPARAM;
		addsz = 0;

		prec  = 0;
		uqval = 0;
		ulval = 0;
		flags = 0;
		sign = '\0';
		base = 0;
		hextable = xdigs_lc;
	rflag:
		pfmt++;
		/* handle various types */
		switch(*pfmt){
		case '%':
			addptr = "%";
			addsz = 1;
			break;
		case '#':
			flags |= FL_ALTNOTA;
			goto rflag;
		case '+':
			sign = '+';
			goto rflag;
		case '0':
			flags |= FL_ZEROPAD;
			goto rflag;
		case '.':
			flags |= FL_RJUSTIFY;
			goto rflag;
		case '=':
			flags |= FL_EXACT;
			goto rflag;
		case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
			prec = 0;
                        do
                                prec = 10 * prec + to_digit(*pfmt++);
                        while (is_digit(*pfmt));
			flags |= FL_HAVEPREC;
			pfmt--;
			goto rflag;
		case 'q':
			flags |= FL_QUADINT;
			goto rflag;
		case 'h':
			flags |= FL_SHRTINT;
			goto rflag;
		case 'l':
			if (flags & FL_LONGINT)
				flags |= FL_QUADINT;
			else
				flags |= FL_LONGINT;
			goto rflag;
		case 'u':
                        if (flags & FL_QUADINT) 
                                uqval = va_arg(ap, quad_t);
                        else
                                ulval = UARG();
                        base = 10;
			goto number;
		case 'D':
			flags |= FL_LONGINT;
 		case 'd':
		case 'i':
                        if (flags & FL_QUADINT) {
                                uqval = va_arg(ap, quad_t);
                                if ((quad_t)uqval < 0) {
                                        uqval = -uqval;
                                        sign = '-';
                                }
                        } else {
                                ulval = SARG();
                                if ((long)ulval < 0) {
                                        ulval = -ulval;
                                        sign = '-';
                                }
                        }
                        base = 10;
			goto number;
		case 'p':
			flags &= ~FL_QUADINT;
			flags |= FL_ALTNOTA;
			goto hexnumber;
 		case 'X':
			hextable = xdigs_uc;
		case 'x':
		hexnumber:
                        if (flags & FL_QUADINT) {
                                uqval = va_arg(ap, quad_t);
                        } else {
                                ulval = UARG();
                        }
			base = 16;
			/* for hex values use exact representation */
			flags |= FL_EXACT;
		number:
                        if (flags & FL_QUADINT)
				addptr = __uqtoa(uqval, sprn + MB_MAXPRNPARAM, base, hextable);
                        else
				addptr = __ultoa(ulval, sprn + MB_MAXPRNPARAM, base, hextable);

			addsz =  sprn + MB_MAXPRNPARAM - addptr;
			if (prec > MB_MAXPRNPARAM)
				prec = MB_MAXPRNPARAM;
			break;
		case 'c':
			flags &= ~FL_EXACT;
/* 			flags &= ~FL_ZEROPAD; */
/* 			flags &= ~FL_HAVEPREC; */
			chval = SARG();
			addptr = &chval;
			addsz = 1;
			break;
		case 's':
			/* for safe operations disable zeropadding and exact*/
			flags &= ~FL_ZEROPAD;
			flags &= ~FL_EXACT;
			addptr = va_arg(ap, char *);
			if (addptr == NULL){
				addptr = "(null)";
				addsz = 6;
			}
			
			if (flags & FL_HAVEPREC){
				register size_t i;
				register char *p;
				p = addptr;
				for (i = 0; i < prec; i++){
					if (p[i] == '\0')
						break;
				}
				if ( i < prec ){
					addsz = prec;
					flags &= ~FL_HAVEPREC;
				} else {
					addsz = i;
				}
			} else
				addsz = strlen(addptr);
			sign='\0';
			break;
		default:
			if (*pfmt == '\0')
				goto done;
			flags &= ~FL_EXACT;
/* 			flags &= ~FL_ZEROPAD; */
/* 			flags &= ~FL_HAVEPREC; */
			chval = *pfmt;
			addptr = &chval;
			addsz = 1;
			break;
		}


		if ((flags & FL_HAVEPREC) != 0){
			if (addsz > prec && (flags & FL_EXACT) != 0){
				if ((flags & (FL_ZEROPAD | FL_RJUSTIFY)) != 0){
					register size_t dlen = addsz - prec;
					addptr += dlen;
					addsz -= dlen;
				} else {
					addsz = prec;
				}
			} else if (prec > addsz){
				register size_t dlen = prec - addsz;
				if ((flags & FL_ZEROPAD) != 0)
					while(dlen--)
						mb_append(mb, "0", 1);
				else if ((flags & FL_RJUSTIFY) != 0)
					while(dlen--)
						mb_append(mb, " ", 1);
			}
		}

		/* do some additional formatting */
		if (base == 16 && (flags & FL_ALTNOTA)){
			addptr -= 2;
			addsz +=  2;
			*addptr = '0';
			*(addptr+1) = 'x';
		}

		if (base == 10 && sign != '\0'){
			addptr--;
			addsz++;
			*addptr = sign;
		}

		err = mb_append(mb, addptr, addsz);
		if (err != 0)
			return err;

		pfmt++;		
		fmt = pfmt;
	}
done:
	return 0;
#undef MB_MAXFMTPARAM
#undef MB_MAXPRNPARAM
}

