/*-
 * Copyright (c) 2001
 *		Stepachev Andrey <aka50@mail.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * strops.c -- String operations
 *
 * $Id: strops.c,v 1.1.1.1 2003/07/06 11:28:37 aka50 Exp $
 */
#include <sys/types.h>
#include <sys/errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#include "dmalloc.h"
#include "strops.h"

/* Trim leading and ending spaces in place */
void 
strtrim (char *str)
{
	long first, last;

	first = 0;
	last = strlen(str);

	if (!last--)
		return;

	// Clip end first
	while (isspace(str[last]) && last > 0)
		last--;
	str[last + 1] = '\0';

	// Clip beginning
	while (isspace(str[first]) && (first < last))
		first++;

	strcpy(str, str + first);
}

void
stripquotes(char *s)
{
	register char *p;
	register char *q;
	register char c;

	if (s == NULL)
		return;

	p = q = s;
	do {
		c = *p++;
		if (c == '\\')
			c = *p++;
		else if (c == '"')
			continue;

		*q++ = c;
	} while (c != '\0');
}

/* Splits line into parts.*/
int
splitline(char *line,char **args,int maxargs,char *delim)
{
	int argcnt = 0;

	char c,s;
	char *d;
	char q = 0;
 
	int qs = 0;

	if (line == NULL || delim == NULL)
		return argcnt;

	for (d = delim; (s = *d++) != 0; ) {
		if (s == '\'' || s == '\"') {
			qs++;
			break;
		}
	}

	/* skip leading delimiters */
cont:
	c = *line++;
	if (c == 0) 
		goto done;

	for (d = delim; (s = *d++) != 0; ){
		/* here we found string in quotes */
		if (qs && (c == '\'' || c == '\"')) {
			q = c;
			c = *line++;
			break;
		}

		if (c == s) 
			goto cont;  /* if ok, try to find next delimiter */
	}

	if (c == 0)
		goto done;

	args[argcnt++] = line-1;
	if (argcnt == maxargs)
		return --argcnt;

	for(;;) {
		c = *line++;
		if (c == 0)
			goto done;

		if (q && c!= q)
			continue; /* check for quote */

		if (q)
			q = 0; /* if we passed throught quote zero it */

		for (d = delim; (s = *d++)!= 0; ) {
			if (c == s) {
				char *w = line - 1;
				*w = '\0';
				goto cont;
			}
		}
	}

	if (q)
		argcnt = 0;

done:
	return argcnt;
}


int 
stracat(char *dst, const char *src, int siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;
	size_t dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (n-- != 0 && *d != '\0')
		d++;

	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return(dlen + strlen(s));

	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
  
	*d = '\0';

	return (dlen + (s - src));  /* count does not include NUL */
}

int
sstrreplace(char *str, int len, char **bufstr, int *buflen, char *what, char *with)
{
	int rcnt = 0;
	char *d, *s, *s2, *p, *p2, *n;

	/* We don't want binary data */
	if (strlen(str) > len) {
		return -1;
	}

	if (bufstr == NULL)
		return -1;

	if (*bufstr)
		DFREE(*bufstr);
  
	/* first, count all replacements */
	p = str;
	s = p + len;
	for (p = str;(p && *p && p<s);p = strstr(p,what)){
		p += strlen(what);
		rcnt++;
	}

	*buflen = len - strlen(what) * rcnt + strlen(with) * rcnt;
	*bufstr = DMALLOC(*buflen + 1, char *);
	if (bufstr == NULL)
		return -1;

	memset(*bufstr,0,*buflen+1);

	p = str;
	s = p+len;

	d = *bufstr;
	s2 = d + *buflen;

	n = strstr(p,what);

	while (p < s && d < s2) {
		if (n) {
			while (p < n && d < s2)
				*(d++) = *(p++);

			p += strlen(what);
			p2 = with;
			while (*p2 && d<s2)
				*(d++) = *(p2++);

			n = strstr(p,what);
		} else {
			*(d++) = *(p++);
		}
	}

	return rcnt;
}

int
strtokcmp(register const char *s1, register const char *s2, const char **last, char *sep)
{
	register char *d;
	register const char *reol;
	register char s;

cont:
	/* Skip leading delimiters */
	for (d = sep; (s = *d++) != 0; ) {
		if (*s1 == s) {
			s1++;
			goto cont;
		}
	}

	reol = s1;
	while (*reol) {
		/* Find real eol */
		for (d = sep; (s = *d++) != 0; ) {
			if (*reol == s)
				goto cont2;
		}
		reol++;
	}

cont2:
	if (last)
		*last = reol;

	while (s1 != reol) {
		if (*s1 != *s2)
			goto done;

		s1++;
		s2++;
	}
done:
	if (*s2 == 0)
		return 0;
	else
		return (*(const unsigned char *)s1 - *(const unsigned char *)s2);

	return (0);
}

/* 
 *       arpa 
 * (from sendmail, modified)
 */
char *
arpadate( register char *ud )
{
	static char b[43 + TZNAME_MAX];
	register char *p;
	register char *q;
	register int i;
	time_t t;

#ifdef ARPA_ADD_TIMEZONE
	register int off;
	register struct tm *lt;
	struct tm gmt;
	char *tz;
#endif

	/*
	 * Get current time.
	 *   This will be used if a null argument is passed and
	 *   to resolve the timezone.
	 */
	t = time(NULL);
	if (ud == NULL)
		ud = ctime(&t);

	/*
	 * Crack the UNIX date line in a singularly unoriginal way.
	 */
	q = b;

	p = &ud[0];		/* Mon */
	*q++= *p++;
	*q++= *p++;
	*q++= *p++;
	*q++= ',';
	*q++= ' ';

	p = &ud[8];		/* 16 */
	if (*p == ' ')
		p++;
	else
		*q++= *p++;

	*q++ = *p++;
	*q++ = ' ';

	p = &ud[4];		/* Sep */
	*q++ = *p++;
	*q++ = *p++;
	*q++ = *p++;
	*q++ = ' ';

	p = &ud[20];		/* 1979 */
	*q++ = *p++;
	*q++ = *p++;
	*q++ = *p++;
	*q++ = *p++;
	*q++ = ' ';

	p = &ud[11];		/* 01:03:52 */
	for (i = 8; i > 0; i--)
		*q++ = *p++;

#ifdef ARPA_ADD_TIMEZONE
	/*
	 * should really get the timezone from the time in "ud" (which
	 * is only different if a non-null arg was passed which is different
	 * from the current time), but for all practical purposes, returning
	 * the current local zone will do (its all that is ever needed).
	 */
	gmt = *gmtime(&t);
	lt = localtime(&t);

	off = (lt->tm_hour - gmt.tm_hour) * 60 + lt->tm_min - gmt.tm_min;

	/* assume that offset isn't more than a day ... */
	if (lt->tm_year < gmt.tm_year)
		off -= 24 * 60;
	else if (lt->tm_year > gmt.tm_year)
		off += 24 * 60;
	else if (lt->tm_yday < gmt.tm_yday)
		off -= 24 * 60;
	else if (lt->tm_yday > gmt.tm_yday)
		off += 24 * 60;

	*q++= ' ';
	if (off == 0) {
		*q++ = 'G';
		*q++ = 'M';
		*q++ = 'T';
	} else {
		tz = lt->tm_zone;
		if (off < 0) {
			off = -off;
			*q++ = '-';
		} else
			*q++= '+';

		if (off >= 24*60)        /* should be impossible */
			*q++ = (off / 600) + '0';

		*q++ = (off / 60) % 10 + '0';
		off %= 60;
		*q++ = (off / 10) + '0';
		*q++ = (off % 10) + '0';
		if (tz != NULL && *tz != '\0') {
			*q++ = ' ';
			*q++ = '(';
			while (*tz != '\0' && q < &b[sizeof b - 3])
				*q++= *tz++;
			*q++ = ')';
		}
	}
#endif
	*q = '\0';

	return b;
}

/*
 * Escape string STE to format:
 *  - all printable chars except ' '(space) and '%' putted as is.
 *  - % as %%
 *  - other in form %<hex-val>(f.i: %20 - space)
 *
 * Return:
 *    0 - all ok (ret points to newly allocated memory)
 *   -1 - error before STE processing (sets *ret to NULL)
 *   -2 - error while STE processing (sets *ret to NULL)
 */
int
strop_escape(char *ste, char **ret)
{
	register unsigned char *cns, *c;
	char *newstr;

#define STROP_CHECKLEN(need) do {				\
		if ((4096 - ((int)cns - (int)newstr)) < need) {	\
			/* no space */				\
			free(newstr);				\
			return -2;				\
		}						\
	} while (0)

	*ret = NULL;
	if (strlen(ste) > 4096)
		/* string is too long */
		return -1;

	newstr = (char *)malloc(4096 + 1);
	if (newstr == NULL)
		return -1;

	for (cns = newstr, c = ste; *c != 0; c++) {
		STROP_CHECKLEN(1);

		if (!isprint(*c) || *c == ' ' || *c == '%') {
			/* not printable or ' ' or '%' */
			*cns++ = '%';

			if (*c == '%') {
				STROP_CHECKLEN(1);
				*cns++ = '%';
			} else {
				register unsigned char nc;
				int i;

				STROP_CHECKLEN(2);

				for (i = 0, nc = *c / 16;
				     i < 2;
				     i++, nc = *c % 16)
				{
					switch (nc) {
					case 15:
						nc = 'F';
						break;
					case 14:
						nc = 'E';
						break;
					case 13:
						nc = 'D';
						break;
					case 12:
						nc = 'C';
						break;
					case 11:
						nc = 'B';
						break;
					case 10:
						nc = 'A';
						break;
					default:
						nc += 48;
						break;
					}
					*cns++ = nc;
				}
			}
		} else {
			/* AS IS */
			*cns++ = *c;
		}
	}
	*cns = '\0';

	*ret = newstr;
	return 0;
#undef STROP_CHECKLEN		
}

/*
 * Unescape string EDS escaped in `strop_escape()' format
 *
 * Return:
 *    -1 - fail (*ret -> NULL)
 *     0 - ok (*ret -> newly allocated memory)
 */
int
strop_unescape(char *eds, char **ret)
{
	register unsigned char *c, *cns;
	char *newstr;

	*ret = NULL;

	newstr = (char*)malloc(strlen(eds));
	if (newstr == NULL)
		return -1;

	for (cns = newstr, c = eds; *c != 0; c++) {
		if (*c == '%') {
			c++;
			if (*c == '%')
				*cns++ = *c;
			else {
				/* non printable */
				*cns = (*c - (*c > 57 ? 55 : 48)) * 16;
				c++;
				*cns += (*c - (*c > 57 ? 55 : 48));
				cns++;
			}
		} else {
			*cns++ = *c;
		}
	}
			
	*ret = newstr;
	return 0;
}
