/*-
 * Copyright (c) 2001
 *		Stepachev Andrey <aka50@mail.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * user.c -- IChat users operations
 * 
 * $Id: user.c,v 1.6 2003/08/31 17:26:12 aka50 Exp $
 */

#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>
#include <string.h>

#include "dmalloc.h"
#include "membuf.h"
#include "queue.h"

#include "ichat.h"
#include "gui.h"
#include "proto.h"
#include "user.h"

static TAILQ_HEAD(, user) userlist;

static struct user *user_new_or_exist(const char *sign, const char *nickname, const char *ver, int state);
static void user_del(const char *sign);
static void user_del2(struct user *usr);

struct user *
find_user(const char *sign)
{
	struct user *n;

	TAILQ_FOREACH(n, &userlist, link){
		if (!strcmp(n->sig, sign))
			return n;
	}
	
	return NULL;
}

struct user *
find_userbynick(const char *nick)
{
	struct user *n;

	TAILQ_FOREACH(n, &userlist, link){
		if (!strcmp(n->nick, nick))
			return n;
	}
	
	return NULL;
}


static struct user *
user_new_or_exist(const char *sign, const char *nickname, const char *ver, int state)
{
	struct user *n;

	n = find_user(sign);
	if (n == NULL){

		n = DTMALLOC(struct user);
		if (n == NULL)
			return n;
		n->sig   = strdup(sign);
		n->nick  = strdup(nickname);
		n->ver   = strdup(ver);
		n->state = state;
		n->cnt   = 0;
		TAILQ_INSERT_TAIL(&userlist, n, link);
	}
	
	return n;
}

static void
user_del(const char *sign)
{
	struct user *n;

	n = find_user(sign);
	if (n == NULL)
		return;

	user_del2(n);
	return;
}

static void
user_del2(struct user *usr)
{
	if (usr == NULL)
		return;

	TAILQ_REMOVE(&userlist, usr, link);
	DFREE(usr->sig);
	DFREE(usr->nick);
	DFREE(usr->ver);
	usr->state = -1;
	DFREE(usr);
	return;
}


int
user_init()
{
	TAILQ_INIT(&userlist);
	return 0;
}

void
user_fini()
{
	struct user *n;

	while (!TAILQ_EMPTY(&userlist)){
		n = TAILQ_FIRST(&userlist);
		user_del2(n);
	}

	return;
}

int
user_lst_update()
{
	struct user *n, *n2;
	time_t curtime;

	curtime = time(NULL);

	n = TAILQ_FIRST(&userlist);
	while(n != NULL){
		n2 = TAILQ_NEXT(n, link);
		if (curtime - n->lseen > ICHAT_USR_TIMEO){
			gui_put_log("%s[%s] timed out...", n->nick, n->sig);
			user_del2(n);
		}
		n = n2;
	}

	return 0;
}

int
user_connect(u_int cnt, 
	     const char *sign, const char *nickname, const char *line, 
	     const char *ver, const char *hmsg, int state)
{
	struct user *n;

	struct membuf pkt, msg;

	mb_init(&pkt);
	mb_init(&msg);

	n = find_user(sign);
	if (n == NULL){
		n = user_new_or_exist(sign, nickname, ver, state);
		if (n == NULL)
			return -1;
		gui_put_log("%s[%s] connected to %s: %s", nickname, sign, line, hmsg);

		ich_cmd_refresh(&msg, line, sign, "Hello there!", 0);
		ich_pkt_make(&pkt, sign, &msg);
		ich_output(&pkt);
	} else if (n->cnt != 0 && n->cnt >= cnt)
		return 0;
	n->cnt = cnt;
	n->lseen = time(NULL);
	return 0;
}

int
user_refresh(u_int cnt, 
	     const char *sign, const char *nickname, const char *line, 
	     const char *ver, const char *hmsg, int state)
{
	struct user *n;

	n = find_user(sign);
	if (n == NULL){
		return user_connect(cnt, sign, nickname, line, ver, hmsg, state);
	}
	if (n->cnt >= cnt)
		return 0;
	n->cnt = cnt;
	if (strcmp(nickname, n->nick)){
		gui_put_log("%s renamed to %s", nickname, n->nick);
		DFREE(n->nick);
		n->nick = strdup(nickname);
	}
	n->lseen = time(NULL);
	return 0;
}


int
user_disconnect(u_int cnt, const char *sign, const char *line)
{
	struct user *n;

	n = find_user(sign);
	if (n == NULL)
		return 0;
	gui_put_log("%s[%s] disconnected from %s", n->nick, n->sig, line);
	user_del(sign);
	return 0;
}

int
user_text(u_int cnt, const char *sign, const char *line, const char *message)
{
	struct user *n;

	n = find_user(sign);
	if (n == NULL){
		gui_put_log("Unknown user %s", sign);
		return 0;
	}
	if (n->cnt >= cnt){
		gui_put_log("User %s have invalid counter", sign);
		return 0;
	}

	n->cnt = cnt;
	n->lseen = time(NULL);
	if (strcmp(line, ICHAT_PRIVATE_LINE))
		gui_put_chat(GUI_MT_NORMAL, n->nick, line, "%s", message);
	else
		gui_put_chat(GUI_MT_NORMAL, n->nick, "MESSAGE", "%s", message);
	return 0;
}

int
user_status(u_int cnt, const char *sign)
{
	struct membuf pkt, msg;

	if (!strcmp(sign, signature))
		return 0;

	mb_init(&pkt);
	mb_init(&msg);

	ich_cmd_status(&msg, "---", status);
	ich_pkt_make(&pkt, sign, &msg);
	ich_output(&pkt);

	mb_free(&pkt);
	mb_free(&msg);

	return 0;
}

int
user_board(u_int cnt, const char *sign, int chunk, const char *text)
{
	struct user *n;

	n = find_user(sign);
	if (n == NULL)
		return 0;
	if (n->cnt >= cnt)
		return 0;

/* 	gui_put_chat(GUI_MT_SYSTEM, n->nick, ICHAT_COMMON_LINE, "Board message %s: %s", n->nick, text); */
	return 0;
}

int
user_recieved(u_int cnt, const char *sign, const char *line, const char *text)
{
	struct user *n;

	n = find_user(sign);
	if (n == NULL)
		return 0;
	if (n->cnt >= cnt)
		return 0;

	n->cnt = cnt;
	n->lseen = time(NULL);
	gui_put_chat(GUI_MT_SYSTEM, n->nick, line, "Message recieved by %s: %s", n->nick, text);
	return 0;
}


/* GUI commands */
int
user_sys_list()
{
	struct user *n;

	gui_put_log("===================");
	gui_put_log("Users:");
	TAILQ_FOREACH(n, &userlist, link){
		gui_put_log("%s [%s]", n->nick, n->sig);
	}
	gui_put_log("===================");
	return 0;
}
